@@ -1,9 +1,4 @@
-use crate::{
-    CompositeField, Constraint, ConstraintAddress, ConstraintAssociatedGroup, ConstraintClose,
-    ConstraintExecutable, ConstraintGroup, ConstraintHasOne, ConstraintInit, ConstraintLiteral,
-    ConstraintMut, ConstraintOwner, ConstraintRaw, ConstraintRentExempt, ConstraintSeedsGroup,
-    ConstraintSigner, ConstraintState, Field, PdaKind, Ty,
-};
+use crate::*;
 use proc_macro2_diagnostics::SpanDiagnosticExt;
 use quote::quote;
 use syn::Expr;
@@ -45,12 +40,10 @@ pub fn generate_composite(f: &CompositeField) -> proc_macro2::TokenStream {
 
 // Linearizes the constraint group so that constraints with dependencies
 // run after those without.
-//
-// The associated cosntraint should always be first since it may also create
-// an account with a PDA.
 pub fn linearize(c_group: &ConstraintGroup) -> Vec<Constraint> {
     let ConstraintGroup {
         init,
+        zeroed,
         mutable,
         signer,
         has_one,
@@ -61,22 +54,21 @@ pub fn linearize(c_group: &ConstraintGroup) -> Vec<Constraint> {
         seeds,
         executable,
         state,
-        associated,
         close,
         address,
     } = c_group.clone();
 
     let mut constraints = Vec::new();
 
-    if let Some(c) = associated {
-        constraints.push(Constraint::AssociatedGroup(c));
-    }
-    if let Some(c) = seeds {
-        constraints.push(Constraint::Seeds(c));
+    if let Some(c) = zeroed {
+        constraints.push(Constraint::Zeroed(c));
     }
     if let Some(c) = init {
         constraints.push(Constraint::Init(c));
     }
+    if let Some(c) = seeds {
+        constraints.push(Constraint::Seeds(c));
+    }
     if let Some(c) = mutable {
         constraints.push(Constraint::Mut(c));
     }
@@ -110,6 +102,7 @@ pub fn linearize(c_group: &ConstraintGroup) -> Vec<Constraint> {
 fn generate_constraint(f: &Field, c: &Constraint) -> proc_macro2::TokenStream {
     match c {
         Constraint::Init(c) => generate_constraint_init(f, c),
+        Constraint::Zeroed(c) => generate_constraint_zeroed(f, c),
         Constraint::Mut(c) => generate_constraint_mut(f, c),
         Constraint::HasOne(c) => generate_constraint_has_one(f, c),
         Constraint::Signer(c) => generate_constraint_signer(f, c),
@@ -120,7 +113,6 @@ fn generate_constraint(f: &Field, c: &Constraint) -> proc_macro2::TokenStream {
         Constraint::Seeds(c) => generate_constraint_seeds(f, c),
         Constraint::Executable(c) => generate_constraint_executable(f, c),
         Constraint::State(c) => generate_constraint_state(f, c),
-        Constraint::AssociatedGroup(c) => generate_constraint_associated(f, c),
         Constraint::Close(c) => generate_constraint_close(f, c),
         Constraint::Address(c) => generate_constraint_address(f, c),
     }
@@ -144,8 +136,28 @@ fn generate_constraint_address(f: &Field, c: &ConstraintAddress) -> proc_macro2:
     }
 }
 
-pub fn generate_constraint_init(_f: &Field, _c: &ConstraintInit) -> proc_macro2::TokenStream {
-    quote! {}
+pub fn generate_constraint_init(f: &Field, c: &ConstraintInitGroup) -> proc_macro2::TokenStream {
+    generate_constraint_init_group(f, c)
+}
+
+pub fn generate_constraint_zeroed(f: &Field, _c: &ConstraintZeroed) -> proc_macro2::TokenStream {
+    let field = &f.ident;
+    let (account_ty, account_wrapper_ty, _) = parse_ty(f);
+    quote! {
+        let #field: #account_wrapper_ty<#account_ty> = {
+            let mut __data: &[u8] = &#field.try_borrow_data()?;
+            let mut __disc_bytes = [0u8; 8];
+            __disc_bytes.copy_from_slice(&__data[..8]);
+            let __discriminator = u64::from_le_bytes(__disc_bytes);
+            if __discriminator != 0 {
+                return Err(anchor_lang::__private::ErrorCode::ConstraintZero.into());
+            }
+            #account_wrapper_ty::try_from_unchecked(
+                program_id,
+                &#field,
+            )?
+        };
+    }
 }
 
 pub fn generate_constraint_close(f: &Field, c: &ConstraintClose) -> proc_macro2::TokenStream {
@@ -186,6 +198,8 @@ pub fn generate_constraint_signer(f: &Field, _c: &ConstraintSigner) -> proc_macr
     let info = match f.ty {
         Ty::AccountInfo => quote! { #ident },
         Ty::ProgramAccount(_) => quote! { #ident.to_account_info() },
+        Ty::Loader(_) => quote! { #ident.to_account_info() },
+        Ty::CpiAccount(_) => quote! { #ident.to_account_info() },
         _ => panic!("Invalid syntax: signer cannot be specified."),
     };
     quote! {
@@ -257,54 +271,53 @@ pub fn generate_constraint_rent_exempt(
     }
 }
 
-pub fn generate_constraint_seeds(f: &Field, c: &ConstraintSeedsGroup) -> proc_macro2::TokenStream {
-    if c.is_init {
-        generate_constraint_seeds_init(f, c)
-    } else {
-        generate_constraint_seeds_address(f, c)
-    }
-}
-
-fn generate_constraint_seeds_init(f: &Field, c: &ConstraintSeedsGroup) -> proc_macro2::TokenStream {
+fn generate_constraint_init_group(f: &Field, c: &ConstraintInitGroup) -> proc_macro2::TokenStream {
     let payer = {
         let p = &c.payer;
         quote! {
             let payer = #p.to_account_info();
         }
     };
-    let seeds_constraint = generate_constraint_seeds_address(f, c);
-    let seeds_with_nonce = {
-        let s = &c.seeds;
-        match c.bump.as_ref() {
-            None => quote! {
-                [#s]
-            },
-            Some(b) => quote! {
-                [#s, &[#b]]
-            },
+
+    let seeds_with_nonce = match &c.seeds {
+        None => quote! {},
+        Some(c) => {
+            let s = &c.seeds;
+            let inner = match c.bump.as_ref() {
+                // Bump target not given. Use the canonical bump.
+                None => {
+                    quote! {
+                        [
+                            #s,
+                            &[
+                                Pubkey::find_program_address(
+                                    &[#s],
+                                    program_id,
+                                ).1
+                            ]
+                        ]
+                    }
+                }
+                // Bump target given. Use it.
+                Some(b) => quote! {
+                    [#s, &[#b]]
+                },
+            };
+            quote! {
+                &#inner[..]
+            }
         }
     };
-    generate_pda(
-        f,
-        seeds_constraint,
-        seeds_with_nonce,
-        payer,
-        &c.space,
-        false,
-        &c.kind,
-    )
+    generate_pda(f, seeds_with_nonce, payer, &c.space, &c.kind)
 }
 
-fn generate_constraint_seeds_address(
-    f: &Field,
-    c: &ConstraintSeedsGroup,
-) -> proc_macro2::TokenStream {
+fn generate_constraint_seeds(f: &Field, c: &ConstraintSeedsGroup) -> proc_macro2::TokenStream {
     let name = &f.ident;
+    let s = &c.seeds;
 
-    // If the bump is provided on *initialization*, then force it to be the
-    // canonical nonce.
+    // If the bump is provided with init *and target*, then force it to be the
+    // canonical bump.
     if c.is_init && c.bump.is_some() {
-        let s = &c.seeds;
         let b = c.bump.as_ref().unwrap();
         quote! {
             let (__program_signer, __bump) = anchor_lang::solana_program::pubkey::Pubkey::find_program_address(
@@ -320,14 +333,22 @@ fn generate_constraint_seeds_address(
         }
     } else {
         let seeds = match c.bump.as_ref() {
+            // Bump target not given. Find it.
             None => {
-                let s = &c.seeds;
                 quote! {
-                    [#s]
+                    [
+                        #s,
+                        &[
+                            Pubkey::find_program_address(
+                                &[#s],
+                                program_id,
+                            ).1
+                        ]
+                    ]
                 }
             }
+            // Bump target given. Use it.
             Some(b) => {
-                let s = &c.seeds;
                 quote! {
                     [#s, &[#b]]
                 }
@@ -335,7 +356,7 @@ fn generate_constraint_seeds_address(
         };
         quote! {
             let __program_signer = Pubkey::create_program_address(
-                &#seeds,
+                &#seeds[..],
                 program_id,
             ).map_err(|_| anchor_lang::__private::ErrorCode::ConstraintSeeds)?;
             if #name.to_account_info().key != &__program_signer {
@@ -345,64 +366,6 @@ fn generate_constraint_seeds_address(
     }
 }
 
-pub fn generate_constraint_associated(
-    f: &Field,
-    c: &ConstraintAssociatedGroup,
-) -> proc_macro2::TokenStream {
-    if c.is_init {
-        generate_constraint_associated_init(f, c)
-    } else {
-        generate_constraint_associated_seeds(f, c)
-    }
-}
-
-pub fn generate_constraint_associated_init(
-    f: &Field,
-    c: &ConstraintAssociatedGroup,
-) -> proc_macro2::TokenStream {
-    let associated_target = c.associated_target.clone();
-    let payer = match &c.payer {
-        None => quote! {
-            let payer = #associated_target.to_account_info();
-        },
-        Some(p) => quote! {
-            let payer = #p.to_account_info();
-        },
-    };
-    let seeds_constraint = generate_constraint_associated_seeds(f, c);
-    let seeds_with_nonce = {
-        if c.associated_seeds.is_empty() {
-            quote! {
-                [
-                    &b"anchor"[..],
-                    #associated_target.to_account_info().key.as_ref(),
-                    &[nonce],
-                ]
-            }
-        } else {
-            let seeds = to_seeds_tts(&c.associated_seeds);
-            quote! {
-                [
-                    &b"anchor"[..],
-                    #associated_target.to_account_info().key.as_ref(),
-                    #seeds
-                    &[nonce],
-                ]
-            }
-        }
-    };
-
-    generate_pda(
-        f,
-        seeds_constraint,
-        seeds_with_nonce,
-        payer,
-        &c.space,
-        true,
-        &c.kind,
-    )
-}
-
 fn parse_ty(f: &Field) -> (proc_macro2::TokenStream, proc_macro2::TokenStream, bool) {
     match &f.ty {
         Ty::ProgramAccount(ty) => {
@@ -454,53 +417,14 @@ fn parse_ty(f: &Field) -> (proc_macro2::TokenStream, proc_macro2::TokenStream, b
 
 pub fn generate_pda(
     f: &Field,
-    seeds_constraint: proc_macro2::TokenStream,
     seeds_with_nonce: proc_macro2::TokenStream,
     payer: proc_macro2::TokenStream,
     space: &Option<Expr>,
-    assign_nonce: bool,
-    kind: &PdaKind,
+    kind: &InitKind,
 ) -> proc_macro2::TokenStream {
     let field = &f.ident;
     let (account_ty, account_wrapper_ty, is_zero_copy) = parse_ty(f);
 
-    let space = match space {
-        // If no explicit space param was given, serialize the type to bytes
-        // and take the length (with +8 for the discriminator.)
-        None => match is_zero_copy {
-            false => {
-                quote! {
-                    let space = 8 + #account_ty::default().try_to_vec().unwrap().len();
-                }
-            }
-            true => {
-                quote! {
-                    let space = 8 + anchor_lang::__private::bytemuck::bytes_of(&#account_ty::default()).len();
-                }
-            }
-        },
-        // Explicit account size given. Use it.
-        Some(s) => quote! {
-            let space = #s;
-        },
-    };
-
-    let nonce_assignment = match assign_nonce {
-        false => quote! {},
-        true => match &f.ty {
-            Ty::CpiAccount(_) => quote! {},
-            _ => match is_zero_copy {
-                false => quote! {
-                    pa.__nonce = nonce;
-                },
-                // Zero copy is not deserialized, so the data must be lazy loaded.
-                true => quote! {
-                    pa.load_init()?.__nonce = nonce;
-                },
-            },
-        },
-    };
-
     let (combined_account_ty, try_from) = match f.ty {
         Ty::AccountInfo => (
             quote! {
@@ -515,7 +439,8 @@ pub fn generate_pda(
                 #account_wrapper_ty<#account_ty>
             },
             quote! {
-                #account_wrapper_ty::try_from_init(
+                #account_wrapper_ty::try_from_unchecked(
+                    program_id,
                     &#field.to_account_info(),
                 )?
             },
@@ -523,75 +448,88 @@ pub fn generate_pda(
     };
 
     match kind {
-        PdaKind::Token { owner, mint } => quote! {
-            let #field: #combined_account_ty = {
-                #space
-                #payer
-                #seeds_constraint
+        InitKind::Token { owner, mint } => {
+            let create_account = generate_create_account(
+                field,
+                quote! {anchor_spl::token::TokenAccount::LEN},
+                quote! {token_program.to_account_info().key},
+                seeds_with_nonce,
+            );
+            quote! {
+                let #field: #combined_account_ty = {
+                    // Define payer variable.
+                    #payer
 
-                // Fund the account for rent exemption.
-                let required_lamports = __anchor_rent
-                    .minimum_balance(anchor_spl::token::TokenAccount::LEN)
-                    .max(1)
-                    .saturating_sub(#field.to_account_info().lamports());
-                if required_lamports > 0 {
-                    anchor_lang::solana_program::program::invoke(
-                        &anchor_lang::solana_program::system_instruction::transfer(
-                            payer.to_account_info().key,
-                            #field.to_account_info().key,
-                            required_lamports,
-                        ),
-                        &[
-                            payer.to_account_info(),
-                            #field.to_account_info(),
-                            system_program.to_account_info().clone(),
-                        ],
-                    )?;
-                }
+                    // Create the account with the system program.
+                    #create_account
 
-                // Allocate space.
-                anchor_lang::solana_program::program::invoke_signed(
-                    &anchor_lang::solana_program::system_instruction::allocate(
-                        #field.to_account_info().key,
-                        anchor_spl::token::TokenAccount::LEN as u64,
-                    ),
-                    &[
-                        #field.to_account_info(),
-                        system_program.clone(),
-                    ],
-                    &[&#seeds_with_nonce[..]],
-                )?;
-
-                // Assign to the spl token program.
-                let __ix = anchor_lang::solana_program::system_instruction::assign(
-                    #field.to_account_info().key,
-                    token_program.to_account_info().key,
-                );
-                anchor_lang::solana_program::program::invoke_signed(
-                    &__ix,
-                    &[
-                        #field.to_account_info(),
-                        system_program.to_account_info(),
-                    ],
-                    &[&#seeds_with_nonce[..]],
-                )?;
-
-                // Initialize the token account.
-                let cpi_program = token_program.to_account_info();
-                let accounts = anchor_spl::token::InitializeAccount {
-                    account: #field.to_account_info(),
-                    mint: #mint.to_account_info(),
-                    authority: #owner.to_account_info(),
-                    rent: rent.to_account_info(),
+                    // Initialize the token account.
+                    let cpi_program = token_program.to_account_info();
+                    let accounts = anchor_spl::token::InitializeAccount {
+                        account: #field.to_account_info(),
+                        mint: #mint.to_account_info(),
+                        authority: #owner.to_account_info(),
+                        rent: rent.to_account_info(),
+                    };
+                    let cpi_ctx = CpiContext::new(cpi_program, accounts);
+                    anchor_spl::token::initialize_account(cpi_ctx)?;
+                    anchor_lang::CpiAccount::try_from_unchecked(
+                        &#field.to_account_info(),
+                    )?
                 };
-                let cpi_ctx = CpiContext::new(cpi_program, accounts);
-                anchor_spl::token::initialize_account(cpi_ctx)?;
-                anchor_lang::CpiAccount::try_from_init(
-                    &#field.to_account_info(),
-                )?
+            }
+        }
+        InitKind::Mint { owner, decimals } => {
+            let create_account = generate_create_account(
+                field,
+                quote! {anchor_spl::token::Mint::LEN},
+                quote! {token_program.to_account_info().key},
+                seeds_with_nonce,
+            );
+            quote! {
+                let #field: #combined_account_ty = {
+                    // Define payer variable.
+                    #payer
+
+                    // Create the account with the system program.
+                    #create_account
+
+                    // Initialize the mint account.
+                    let cpi_program = token_program.to_account_info();
+                    let accounts = anchor_spl::token::InitializeMint {
+                        mint: #field.to_account_info(),
+                        rent: rent.to_account_info(),
+                    };
+                    let cpi_ctx = CpiContext::new(cpi_program, accounts);
+                    anchor_spl::token::initialize_mint(cpi_ctx, #decimals, &#owner.to_account_info().key, None)?;
+                    anchor_lang::CpiAccount::try_from_unchecked(
+                        &#field.to_account_info(),
+                    )?
+                };
+            }
+        }
+        InitKind::Program { owner } => {
+            let space = match space {
+                // If no explicit space param was given, serialize the type to bytes
+                // and take the length (with +8 for the discriminator.)
+                None => match is_zero_copy {
+                    false => {
+                        quote! {
+                            let space = 8 + #account_ty::default().try_to_vec().unwrap().len();
+                        }
+                    }
+                    true => {
+                        quote! {
+                            let space = 8 + anchor_lang::__private::bytemuck::bytes_of(&#account_ty::default()).len();
+                        }
+                    }
+                },
+                // Explicit account size given. Use it.
+                Some(s) => quote! {
+                    let space = #s;
+                },
             };
-        },
-        PdaKind::Program { owner } => {
+
             // Owner of the account being created. If not specified,
             // default to the currently executing program.
             let owner = match owner {
@@ -602,40 +540,14 @@ pub fn generate_pda(
                     &#o
                 },
             };
+            let create_account =
+                generate_create_account(field, quote! {space}, owner, seeds_with_nonce);
             quote! {
                 let #field = {
                     #space
                     #payer
-                    #seeds_constraint
-
-                    let lamports = __anchor_rent.minimum_balance(space);
-                    let ix = anchor_lang::solana_program::system_instruction::create_account(
-                        payer.to_account_info().key,
-                        #field.to_account_info().key,
-                        lamports,
-                        space as u64,
-                        #owner,
-                    );
-
-                    anchor_lang::solana_program::program::invoke_signed(
-                        &ix,
-                        &[
-
-                            #field.to_account_info(),
-                            payer.to_account_info(),
-                            system_program.to_account_info(),
-                        ],
-                        &[&#seeds_with_nonce[..]]
-                    ).map_err(|e| {
-                        anchor_lang::solana_program::msg!("Unable to create associated account");
-                        e
-                    })?;
-
-                    // For now, we assume all accounts created with the `associated`
-                    // attribute have a `nonce` field in their account.
+                    #create_account
                     let mut pa: #combined_account_ty = #try_from;
-
-                    #nonce_assignment
                     pa
                 };
             }
@@ -643,58 +555,85 @@ pub fn generate_pda(
     }
 }
 
-pub fn generate_constraint_associated_seeds(
-    f: &Field,
-    c: &ConstraintAssociatedGroup,
+// Generated code to create an account with with system program with the
+// given `space` amount of data, owned by `owner`.
+//
+// `seeds_with_nonce` should be given for creating PDAs. Otherwise it's an
+// empty stream.
+pub fn generate_create_account(
+    field: &Ident,
+    space: proc_macro2::TokenStream,
+    owner: proc_macro2::TokenStream,
+    seeds_with_nonce: proc_macro2::TokenStream,
 ) -> proc_macro2::TokenStream {
-    let field = &f.ident;
-    let associated_target = c.associated_target.clone();
-    let seeds_no_nonce = if c.associated_seeds.is_empty() {
-        quote! {
-            &b"anchor"[..],
-            #associated_target.to_account_info().key.as_ref(),
-        }
-    } else {
-        let seeds = to_seeds_tts(&c.associated_seeds);
-        quote! {
-            &b"anchor"[..],
-            #associated_target.to_account_info().key.as_ref(),
-            #seeds
-        }
-    };
-
-    let is_find_nonce = match &f.ty {
-        Ty::CpiAccount(_) => true,
-        Ty::AccountInfo => true,
-        _ => c.is_init,
-    };
-    let associated_field = if is_find_nonce {
-        quote! {
-            let (__associated_field, nonce) = Pubkey::find_program_address(
-                &[#seeds_no_nonce],
-                program_id,
-            );
-        }
-    } else {
-        let nonce = match &f.ty {
-            Ty::ProgramAccount(_) => quote! { #field.__nonce },
-            Ty::Loader(_) => {
-                // Zero copy is not deserialized, so the data must be lazy loaded.
-                quote! { #field.load()?.__nonce }
-            }
-            _ => panic!("Invalid type for initializing a program derived address"),
-        };
-        quote! {
-            let __associated_field = Pubkey::create_program_address(
-                &[#seeds_no_nonce &[#nonce]],
-                program_id,
-            )?;
-        }
-    };
     quote! {
-        #associated_field
-        if &__associated_field != #field.to_account_info().key {
-            return Err(anchor_lang::__private::ErrorCode::ConstraintAssociatedInit.into());
+        // If the account being initialized already has lamports, then
+        // return them all back to the payer so that the account has
+        // zero lamports when the system program's create instruction
+        // is eventually called.
+        let __current_lamports = #field.to_account_info().lamports();
+        if __current_lamports == 0 {
+            // Create the token account with right amount of lamports and space, and the correct owner.
+            let lamports = __anchor_rent.minimum_balance(#space);
+            anchor_lang::solana_program::program::invoke_signed(
+                &anchor_lang::solana_program::system_instruction::create_account(
+                    payer.to_account_info().key,
+                    #field.to_account_info().key,
+                    lamports,
+                    #space as u64,
+                    #owner,
+                ),
+                &[
+                    payer.to_account_info(),
+                    #field.to_account_info(),
+                    system_program.to_account_info().clone(),
+                ],
+                &[#seeds_with_nonce],
+            )?;
+        } else {
+            // Fund the account for rent exemption.
+            let required_lamports = __anchor_rent
+                .minimum_balance(#space)
+                .max(1)
+                .saturating_sub(__current_lamports);
+            if required_lamports > 0 {
+                anchor_lang::solana_program::program::invoke(
+                    &anchor_lang::solana_program::system_instruction::transfer(
+                        payer.to_account_info().key,
+                        #field.to_account_info().key,
+                        required_lamports,
+                    ),
+                    &[
+                        payer.to_account_info(),
+                        #field.to_account_info(),
+                        system_program.to_account_info().clone(),
+                    ],
+                )?;
+            }
+            // Allocate space.
+            anchor_lang::solana_program::program::invoke_signed(
+                &anchor_lang::solana_program::system_instruction::allocate(
+                    #field.to_account_info().key,
+                    #space as u64,
+                ),
+                &[
+                    #field.to_account_info(),
+                    system_program.clone(),
+                ],
+                &[#seeds_with_nonce],
+            )?;
+            // Assign to the spl token program.
+            anchor_lang::solana_program::program::invoke_signed(
+                &anchor_lang::solana_program::system_instruction::assign(
+                    #field.to_account_info().key,
+                    #owner,
+                ),
+                &[
+                    #field.to_account_info(),
+                    system_program.to_account_info(),
+                ],
+                &[#seeds_with_nonce],
+            )?;
         }
     }
 }
@@ -729,30 +668,3 @@ pub fn generate_constraint_state(f: &Field, c: &ConstraintState) -> proc_macro2:
         }
     }
 }
-
-// Returns the inner part of the seeds slice as a token stream.
-fn to_seeds_tts(seeds: &[syn::Expr]) -> proc_macro2::TokenStream {
-    assert!(!seeds.is_empty());
-    let seed_0 = &seeds[0];
-    let mut tts = match seed_0 {
-        syn::Expr::Path(_) => quote! {
-            anchor_lang::Key::key(&#seed_0).as_ref(),
-        },
-        _ => quote! {
-            #seed_0,
-        },
-    };
-    for seed in &seeds[1..] {
-        tts = match seed {
-            syn::Expr::Path(_) => quote! {
-                #tts
-                anchor_lang::Key::key(&#seed).as_ref(),
-            },
-            _ => quote! {
-                #tts
-                #seed,
-            },
-        };
-    }
-    tts
-}
