@@ -30,10 +30,10 @@ pub fn generate(accs: &AccountsStruct) -> proc_macro2::TokenStream {
                     }
                 }
                 AccountField::Field(f) => {
-                    // Associated fields are *first* deserialized into
-                    // AccountInfos, and then later deserialized into
-                    // ProgramAccounts in the "constraint check" phase.
-                    if is_pda_init(af) {
+                    // `init` and `zero` acccounts are special cased as they are
+                    // deserialized by constraints. Here, we just take out the
+                    // AccountInfo for later use at constraint validation time.
+                    if is_init(af) || f.constraints.zeroed.is_some() {
                         let name = &f.ident;
                         quote!{
                             let #name = &accounts[0];
@@ -41,17 +41,10 @@ pub fn generate(accs: &AccountsStruct) -> proc_macro2::TokenStream {
                         }
                     } else {
                         let name = typed_ident(f);
-                        match f.constraints.is_init() {
-                            false => quote! {
-                                #[cfg(feature = "anchor-debug")]
-                                ::solana_program::log::sol_log(stringify!(#name));
-                                let #name = anchor_lang::Accounts::try_accounts(program_id, accounts, ix_data)?;
-                            },
-                            true => quote! {
-                                #[cfg(feature = "anchor-debug")]
-                                ::solana_program::log::sol_log(stringify!(#name));
-                                let #name = anchor_lang::AccountsInit::try_accounts_init(program_id, accounts)?;
-                            },
+                        quote! {
+                            #[cfg(feature = "anchor-debug")]
+                            ::solana_program::log::sol_log(stringify!(#name));
+                            let #name = anchor_lang::Accounts::try_accounts(program_id, accounts, ix_data)?;
                         }
                     }
                 }
@@ -114,25 +107,6 @@ pub fn generate(accs: &AccountsStruct) -> proc_macro2::TokenStream {
     }
 }
 
-// Returns true if the given AccountField has an associated init constraint.
-fn is_pda_init(af: &AccountField) -> bool {
-    match af {
-        AccountField::CompositeField(_s) => false,
-        AccountField::Field(f) => {
-            f.constraints
-                .associated
-                .as_ref()
-                .map(|f| f.is_init)
-                .unwrap_or(false)
-                || f.constraints
-                    .seeds
-                    .as_ref()
-                    .map(|f| f.is_init)
-                    .unwrap_or(false)
-        }
-    }
-}
-
 fn typed_ident(field: &Field) -> TokenStream {
     let name = &field.ident;
 
@@ -193,18 +167,17 @@ fn typed_ident(field: &Field) -> TokenStream {
 }
 
 pub fn generate_constraints(accs: &AccountsStruct) -> proc_macro2::TokenStream {
-    // All fields without an `#[account(associated)]` attribute.
-    let non_associated_fields: Vec<&AccountField> =
-        accs.fields.iter().filter(|af| !is_pda_init(af)).collect();
+    let non_init_fields: Vec<&AccountField> =
+        accs.fields.iter().filter(|af| !is_init(af)).collect();
 
-    // Deserialization for each *associated* field. This must be after
+    // Deserialization for each pda init field. This must be after
     // the inital extraction from the accounts slice and before access_checks.
-    let init_associated_fields: Vec<proc_macro2::TokenStream> = accs
+    let init_fields: Vec<proc_macro2::TokenStream> = accs
         .fields
         .iter()
         .filter_map(|af| match af {
             AccountField::CompositeField(_s) => None,
-            AccountField::Field(f) => match is_pda_init(af) {
+            AccountField::Field(f) => match is_init(af) {
                 false => None,
                 true => Some(f),
             },
@@ -213,7 +186,7 @@ pub fn generate_constraints(accs: &AccountsStruct) -> proc_macro2::TokenStream {
         .collect();
 
     // Constraint checks for each account fields.
-    let access_checks: Vec<proc_macro2::TokenStream> = non_associated_fields
+    let access_checks: Vec<proc_macro2::TokenStream> = non_init_fields
         .iter()
         .map(|af: &&AccountField| match af {
             AccountField::Field(f) => constraints::generate(f),
@@ -222,7 +195,7 @@ pub fn generate_constraints(accs: &AccountsStruct) -> proc_macro2::TokenStream {
         .collect();
 
     quote! {
-        #(#init_associated_fields)*
+        #(#init_fields)*
         #(#access_checks)*
     }
 }
@@ -250,3 +223,10 @@ pub fn generate_accounts_instance(accs: &AccountsStruct) -> proc_macro2::TokenSt
         }
     }
 }
+
+fn is_init(af: &AccountField) -> bool {
+    match af {
+        AccountField::CompositeField(_s) => false,
+        AccountField::Field(f) => f.constraints.init.is_some(),
+    }
+}
