@@ -262,7 +262,8 @@ pub struct ErrorCode {
 // All well formed constraints on a single `Accounts` field.
 #[derive(Debug, Default, Clone)]
 pub struct ConstraintGroup {
-    init: Option<ConstraintInit>,
+    init: Option<ConstraintInitGroup>,
+    zeroed: Option<ConstraintZeroed>,
     mutable: Option<ConstraintMut>,
     signer: Option<ConstraintSigner>,
     owner: Option<ConstraintOwner>,
@@ -270,7 +271,6 @@ pub struct ConstraintGroup {
     seeds: Option<ConstraintSeedsGroup>,
     executable: Option<ConstraintExecutable>,
     state: Option<ConstraintState>,
-    associated: Option<ConstraintAssociatedGroup>,
     has_one: Vec<ConstraintHasOne>,
     literal: Vec<ConstraintLiteral>,
     raw: Vec<ConstraintRaw>,
@@ -279,8 +279,8 @@ pub struct ConstraintGroup {
 }
 
 impl ConstraintGroup {
-    pub fn is_init(&self) -> bool {
-        self.init.is_some()
+    pub fn is_zeroed(&self) -> bool {
+        self.zeroed.is_some()
     }
 
     pub fn is_mutable(&self) -> bool {
@@ -297,12 +297,13 @@ impl ConstraintGroup {
 }
 
 // A single account constraint *after* merging all tokens into a well formed
-// constraint. Some constraints like "associated" are defined by multiple
+// constraint. Some constraints like "seeds" are defined by multiple
 // tokens, so a merging phase is required.
 #[allow(clippy::large_enum_variant)]
 #[derive(Debug)]
 pub enum Constraint {
-    Init(ConstraintInit),
+    Init(ConstraintInitGroup),
+    Zeroed(ConstraintZeroed),
     Mut(ConstraintMut),
     Signer(ConstraintSigner),
     HasOne(ConstraintHasOne),
@@ -313,7 +314,6 @@ pub enum Constraint {
     Seeds(ConstraintSeedsGroup),
     Executable(ConstraintExecutable),
     State(ConstraintState),
-    AssociatedGroup(ConstraintAssociatedGroup),
     Close(ConstraintClose),
     Address(ConstraintAddress),
 }
@@ -323,6 +323,7 @@ pub enum Constraint {
 #[derive(Debug)]
 pub enum ConstraintToken {
     Init(Context<ConstraintInit>),
+    Zeroed(Context<ConstraintZeroed>),
     Mut(Context<ConstraintMut>),
     Signer(Context<ConstraintSigner>),
     HasOne(Context<ConstraintHasOne>),
@@ -334,13 +335,13 @@ pub enum ConstraintToken {
     Executable(Context<ConstraintExecutable>),
     State(Context<ConstraintState>),
     Close(Context<ConstraintClose>),
-    Associated(Context<ConstraintAssociated>),
-    AssociatedPayer(Context<ConstraintAssociatedPayer>),
-    AssociatedSpace(Context<ConstraintAssociatedSpace>),
-    AssociatedWith(Context<ConstraintAssociatedWith>),
+    Payer(Context<ConstraintPayer>),
+    Space(Context<ConstraintSpace>),
     Address(Context<ConstraintAddress>),
     TokenMint(Context<ConstraintTokenMint>),
     TokenAuthority(Context<ConstraintTokenAuthority>),
+    MintAuthority(Context<ConstraintMintAuthority>),
+    MintDecimals(Context<ConstraintMintDecimals>),
     Bump(Context<ConstraintTokenBump>),
 }
 
@@ -353,6 +354,9 @@ impl Parse for ConstraintToken {
 #[derive(Debug, Clone)]
 pub struct ConstraintInit {}
 
+#[derive(Debug, Clone)]
+pub struct ConstraintZeroed {}
+
 #[derive(Debug, Clone)]
 pub struct ConstraintMut {}
 
@@ -390,14 +394,19 @@ pub enum ConstraintRentExempt {
     Skip,
 }
 
+#[derive(Debug, Clone)]
+pub struct ConstraintInitGroup {
+    pub seeds: Option<ConstraintSeedsGroup>,
+    pub payer: Option<Ident>,
+    pub space: Option<Expr>,
+    pub kind: InitKind,
+}
+
 #[derive(Debug, Clone)]
 pub struct ConstraintSeedsGroup {
     pub is_init: bool,
     pub seeds: Punctuated<Expr, Token![,]>,
-    pub payer: Option<Ident>,
-    pub space: Option<Expr>,
-    pub kind: PdaKind,
-    pub bump: Option<Expr>,
+    pub bump: Option<Expr>, // None => bump was given without a target.
 }
 
 #[derive(Debug, Clone)]
@@ -414,40 +423,23 @@ pub struct ConstraintState {
 }
 
 #[derive(Debug, Clone)]
-pub struct ConstraintAssociatedGroup {
-    pub is_init: bool,
-    pub associated_target: Expr,
-    pub associated_seeds: Vec<Expr>,
-    pub payer: Option<Ident>,
-    pub space: Option<Expr>,
-    pub kind: PdaKind,
-}
-
-#[derive(Debug, Clone)]
-pub struct ConstraintAssociated {
-    pub target: Expr,
-}
-
-#[derive(Debug, Clone)]
-pub struct ConstraintAssociatedPayer {
+pub struct ConstraintPayer {
     pub target: Ident,
 }
 
 #[derive(Debug, Clone)]
-pub struct ConstraintAssociatedWith {
-    pub target: Expr,
-}
-
-#[derive(Debug, Clone)]
-pub struct ConstraintAssociatedSpace {
+pub struct ConstraintSpace {
     pub space: Expr,
 }
 
 #[derive(Debug, Clone)]
 #[allow(clippy::large_enum_variant)]
-pub enum PdaKind {
+pub enum InitKind {
     Program { owner: Option<Expr> },
+    // Owner for token and mint represents the authority. Not to be confused
+    // with the owner of the AccountInfo.
     Token { owner: Expr, mint: Expr },
+    Mint { owner: Expr, decimals: Expr },
 }
 
 #[derive(Debug, Clone)]
@@ -465,9 +457,19 @@ pub struct ConstraintTokenAuthority {
     auth: Expr,
 }
 
+#[derive(Debug, Clone)]
+pub struct ConstraintMintAuthority {
+    mint_auth: Expr,
+}
+
+#[derive(Debug, Clone)]
+pub struct ConstraintMintDecimals {
+    decimals: Expr,
+}
+
 #[derive(Debug, Clone)]
 pub struct ConstraintTokenBump {
-    bump: Expr,
+    bump: Option<Expr>,
 }
 
 // Syntaxt context object for preserving metadata about the inner item.
